"""
Normalizes an XML DTD by parsing it and writing it back out.
"""

import string, sys
from xml.parsers.xmlproc import dtdparser, xmlapp

# --- Utilities

def content_particle_to_string(cp):
    if len(cp) == 2:
        return cp[0] + cp[1]
    else:
        return content_model_to_string(cp)

def content_model_to_string(cm):
    if type(cm) == type(""):
        return cm

    (sep, cont, mod) = cm
    return "(%s)%s" % (string.join(map(content_particle_to_string,
                                       cont), sep), mod)


def atype_to_string(atype):
    if type(atype) == type([]):
        return "(%s)" % string.join(atype, " | ")
    else:
        return atype

def adecl_to_string(adecl):
    if adecl == "#DEFAULT":
        return ""
    else:
        return adecl

def adef_to_string(adef):
    if adef == None:
        return ""
    else:
        return "'%s'" % attr_escape(adef)


def entity_escape(value):
    return string.replace(string.replace(value, "&", "&amp;"),
                          "'", "&apos;")

def attr_escape(value):
    return string.replace(string.replace(string.replace(value,
                                                        "&", "&amp;"),
                                         "'", "&apos;"), "<", "&lt;")


# --- The DTD application

class DTDWriter(xmlapp.DTDConsumer):

    def __init__(self, outf):
        self._out = outf
        self._attlist = []


    def new_general_entity(self, name, val):
        self._empty_attlist()
        self._out.write("<!ENTITY %s '%s'>\n" %
                        (name, entity_escape(val)))

    def new_external_entity(self, name, pubid, sysid, ndata):
        self._empty_attlist()
        if pubid:
            self._out.write("<!ENTITY %s PUBLIC '%s' '%s'" %
                            (name, pubid, sysid))
        else:
            self._out.write("<!ENTITY %s SYSTEM '%s'" % (name, sysid))

        if ndata:
            self._out.write(" NDATA %s>\n" % ndata)
        else:
            self._out.write(">\n")


    def new_parameter_entity(self, name, val):
        self._empty_attlist()
        self._out.write("<!ENTITY %% %s '%s'>\n" %
                        (name, entity_escape(val)))
    
    def new_external_pe(self, name, pubid, sysid):
        self._empty_attlist()
        if pubid:
            self._out.write("<!ENTITY %% %s PUBLIC '%s' '%s'>\n" %
                            (name, pubid, sysid))
        else:
            self._out.write("<!ENTITY %% %s SYSTEM '%s'>\n" %
                            (name, sysid))


    def new_notation(self, name, pubid, sysid):
        self._empty_attlist()
        if pubid and sysid:
            self._out.write("<!NOTATION name PUBLIC '%s' '%s'>\n" %
                            (name, pubid, sysid))
        elif pubid:
            self._out.write("<!NOTATION name PUBLIC '%s'>\n" %
                            (name, pubid))
        else:
            self._out.write("<!NOTATION name SYSTEM '%s'>\n" %
                            (name, sysid))

    def new_element_type(self, name, cont):
        self._empty_attlist()
        self._out.write("<!ELEMENT %s %s>\n" %
                        (name, content_model_to_string(cont)))
	    
    def new_attribute(self, elem, attr, a_type, a_decl, a_def):
        self._attlist.append((elem, attr, a_type, a_decl, a_def))

    def handle_comment(self, contents):
        self._empty_attlist()
        self._out.write("<!-- %s -->\n" % contents)

    def handle_pi(self, target, data):
        self._empty_attlist()
        self._out.write("<?%s %s?>" % (target, data))


    # internal methods

    def _empty_attlist(self):
        prev = None
        for (elem, attr, a_type, a_decl, a_def) in self._attlist:
            if elem != prev:
                if prev != None:
                    self._out.write(">\n")

                self._out.write("<!ATTLIST %s " % elem)
                prev = elem

            self._out.write("\n          %s %s %s %s" %
                            (attr, atype_to_string(a_type),
                             adecl_to_string(a_decl),
                             adef_to_string(a_def)))

        if self._attlist != []:
            self._out.write(">\n")
            self._attlist = []


# --- Main functions

def normalize_dtd_file(dtdfile, outfile):
    outf = open(outfile, "w")
    
    parser = dtdparser.DTDParser()
    parser.set_dtd_consumer(DTDWriter(outf))
    parser.parse_resource(dtdfile)

    outf.close()

    
# --- Main program

if __name__ == "__main__":
    normalize_dtd_file(sys.argv[1], sys.argv[2])
