"""
An XInclude implementation written as a SAX 2.0 parser filter. The
implementation follows the 2000-03-22 working draft.

Spec: <URL: http://www.w3.org/TR/2000/WD-xinclude-20000322 >

$Id: xinclude.py,v 1.4 2001/08/19 19:27:41 larsga Exp $
"""

import urlparse, urllib

from xml.sax.handler import feature_namespaces
from xml.sax._exceptions import SAXNotSupportedException, SAXException
from xml.sax.saxutils import XMLFilterBase

# --- Global attributes

xinclude_ns = "http://www.w3.org/1999/XML/xinclude"


# --- The filter

class XIncludeFilter(XMLFilterBase):

    def __init__(self, parent, parent_factory):
        """The parent_factory is a callable object that, when called,
        returns an object that can be used as a parent event source
        for the filter."""
        XMLFilterBase.__init__(self, parent)
        self._factory   = parent_factory
        self._xinclude  = (xinclude_ns, "include")
        self._locator   = None
        self._stack     = []
        self._open_uris = []

        parent.setFeature(feature_namespaces, 1)


    # SAX 2.0 methods

    def setDocumentLocator(self, locator):
        self._locator = locator
        self._cont_handler.setDocumentLocator(locator)


    def setFeature(self, feature, state):
        if feature == feature_namespaces and not state:
            raise SAXNotSupportedException("XInclude processing requires "
                                           "namespace processing")

        self._parent.setFeature(feature, state)


    def startDocument(self):
        if self._locator == None:
            raise SAXException("Locator support is required.")
                
        if self._stack == []: # only report the first...
	    self._open_uris.append(self._locator.getSystemId())
            self._cont_handler.startDocument()

    def endDocument(self):
        if self._stack == []: # only report the last...
            self._cont_handler.endDocument()       


    def startElementNS(self, name, qname, attrs):
        if name == self._xinclude:
            try:
                url = self._resolve_absolute_uri(attrs[(None, "href")])
            except KeyError, e:
                self._report_error("The XInclude href attribute is "
                                   "required.")
                return
                
            parse = attrs.get((None, "parse"), "xml")
            if parse == "text":
                self._read_text_from_uri(url)
            elif parse == "xml":
                self._read_xml_from_uri(url)
            else:
                self._report_error("Unknown XInclude parse value '%s'"
                                   % parse)
                
        else:
            self._cont_handler.startElementNS(name, qname, attrs)


    def endElementNS(self, name, qname):
        if name != self._xinclude:
            self._cont_handler.endElementNS(name, qname)


    # XIncludeFilter methods

    def _resolve_absolute_uri(self, uri):
        return urlparse.urljoin(self._locator.getSystemId(), uri)
        
    def _read_text_from_uri(self, uri):
        inf = urllib.urlopen(uri)
        buf = inf.read(16384)
        while buf != "":
            self._cont_handler.characters(buf)
            buf = inf.read(16384)
        inf.close()


    def _report_error(self, message):
        self._err_handler.error(SAXParseException(message, None,
                                                  self._locator))

        
    def _read_xml_from_uri(self, uri):
        if uri in self._open_uris:
            self._report_error("XInclude inclusion loop detected! Bad "
                               "reference to %s." % repr(uri))
            return
        
        self._stack.append((self._parent, self._locator))
        self._open_uris.append(uri)

        self._locator = None # to make sure that it is reset
        self._parent = self._factory()
        self._parent.setContentHandler(self)
        self._parent.setErrorHandler(self)
        self._parent.setEntityResolver(self)
        self._parent.setDTDHandler(self)
        self._parent.setFeature(feature_namespaces, 1)
        self._parent.parse(uri)

        (self._parent, self._locator) = self._stack[-1]
        del self._stack[-1]
        del self._open_uris[-1]


# --- Test program

if __name__ == "__main__":

    import sys
    from xml.sax.saxutils import XMLGenerator
    from xml.sax import make_parser

    f = XIncludeFilter(make_parser(), make_parser)
    f.setContentHandler(ContentGenerator())
    f.parse(sys.argv[1])

